import struct
from math import log2

import matplotlib.pyplot as plt


def get_data(file):
    with open(file, "rb") as f:
        content = f.read()
        data = struct.unpack(f"{len(content)//8}d", content)
    return data

def plot_mem_data(data, title="Memory Usage", xlabel="Time", ylabel="Usage"):
    units = ["B", "KB", "MB", "GB"]
    if len(data) == 0:
        avg = 0
    else:
        avg = sum(data)/len(data)
    if avg == 0:
        i = 0
    else:
        i = int(min(log2(avg)/10, 3))
        
    data = list(map(lambda x: x/2**(i*10), data))
    plt.plot(data)
    plt.title(title)
    plt.xlabel(xlabel)
    plt.ylabel(f"{ylabel} ({units[i]})")
    
def save_plot(data, file):
    plot_mem_data(data)
    plt.savefig(file)

if __name__ == "__main__":
    file = input()
    d = get_data(file)
    plot_mem_data(d)
    plt.show()
    save_plot(d, "plot.png")



from math import log2


class Size:
    def __init__(self, value):
        """"Represents the size of an object

        Args:
            value (int): value in bytes
        """
        
        self._bytes = value
        
    def __repr__(self):
        if self._bytes == 0:
            i = 0
        else:
            i = int(log2(self.bytes)/10)
        if i == 0:
            return f"{round(self.bytes, 1)}B"
        elif i == 1:
            return f"{round(self.kbytes, 1)}KB"
        elif i == 2:
            return f"{round(self.mbytes, 1)}MB"
        else:
            return f"{round(self.gbytes, 1)}GB"
        
    @property
    def bytes(self):
        return self._bytes
    
    @property
    def kbytes(self):
        return self.bytes / 2**10
    
    @property
    def mbytes(self):
        return self.kbytes / 2**10
    
    @property
    def gbytes(self):
        return self.mbytes / 2**10
    
    
class Time:
    def __init__(self, value):
        """Represents a measure of time

        Args:
            value (datetime.timedelta): Time object
        """
        
        self._value = value
        
    def __repr__(self):
        if self._value == 0:
            return "0s"
        if self.years > 0.5:
            return f"{round(self.years)}Y"
        elif self.months > 0.5:
            return f"{round(self.months)}M"
        elif self.days > 0:
            return f"{self.days}D"
        elif self.hours > 0.5:
            return f"{round(self.hours)}h"
        elif self.minutes > 0.5:
            return f"{round(self.minutes)}m"
        else:
            return f"{round(self.seconds)}s"
        
    @property
    def seconds(self):
        return self._value.total_seconds()
    
    @property 
    def minutes(self):
        return self.seconds / 60
    
    @property
    def hours(self):
        return self.minutes / 60
    
    @property
    def days(self):
        if hasattr(self._value, "days"):
            return self._value.days
        return 0
    
    @property
    def months(self):
        return self.days / 30.417
    
    @property
    def years(self):
        return self.days / 365.25



import argparse
import socket
import struct

import termtables as tt
from colorama import Fore, Style

from . import constants as const
from .process import Process
from .units import Size

commands = [
    "init",
    "start",
    "stop",
    "add",
    "rem",
    "kill",
    "restart",
    "status",
    "list",
    "start",
    "monit"
]
commands.sort()

help_text = f"""\
Usage: python -m pypm CMD [OPTIONS]

Positional arguments:
    CMD             {', '.join(commands)}
    
For additional help use python -m pypm CMD --help
"""

def color(text, color):
    """Adds color to given text."""
    return f"{color}{text}{Style.RESET_ALL}"

def isdata(bytearr):
    """Returns True if the given data is an encoded string or binary data"""
    return bytearr[0] == const.DATA_CODE[0]

def get_start_parser():
    parser = argparse.ArgumentParser(prog="python -m pypm init")
    parser.add_argument("--port", 
                        type=int, 
                        default=8080, 
                        help="Network port")
    parser.add_argument("--logdir", 
                        type=str, 
                        default="logs", 
                        help="Logging directory")
    parser.add_argument("--logfreq", 
                        type=int, 
                        default=30, 
                        help="Logging frequency (per minute)")
    return parser

def get_cmd_parser(cmd):
    parser = argparse.ArgumentParser(prog="python -m pypm", 
                                     usage=f"usage: python -m pypm {cmd} \
[-h] [--port PORT] [--host HOST] [ARGS [ARGS ...]]")
    parser.add_argument("args",
                        metavar="ARGS",
                        nargs="*",
                        help="Command arguments")
    parser.add_argument("--port", 
                        type=int, 
                        default=8080, 
                        help="Network port")
    parser.add_argument("--host", 
                        type=str, 
                        default="localhost", 
                        help="Host")
    return parser

def print_msg(text):
    """Prints the given text, coloring it based on the first word"""
    if text.startswith("Error:"):
        print(color(text, Fore.RED))
    elif text.startswith("Warning:"):
        print(color(text, Fore.YELLOW))
    else:
        print(text)

def process_command(cmd, args, host, port):
    """Processes a given command

    Args:
        cmd (str): Command
        args (list): List of command arguments
        host (str): Remote host to connect to
        port (int): Network port
    """
    try:
        if cmd == "stop":
            if len(args) != 0:
                print_msg("Error: this command takes no arguments")
                return
            process_stop_command(args, host, port)
        elif cmd == "add":
            if len(args) < 2:
                print_msg("Error: Not enough arguments (need at least NAME and COMMAND)")
                return
            if len(args) > 4:
                print_msg("Error: Too many arguments")
            process_add_command(args, host, port)
        elif cmd == "start":
            if len(args) > 1:
                print_msg("Error: Invalid number of arguments")
                return
            process_start_command(args, host, port)
        elif cmd == "restart":
            if len(args) > 1:
                print_msg("Error: Invalid number of arguments")
                return
            process_restart_command(args, host, port)
        elif cmd == "rem":
            if len(args) != 1:
                print_msg("Error: Invalid number of arguments")
                return
            process_remove_command(args, host, port)
        elif cmd == "kill":
            if len(args) != 1:
                print_msg("Error: Invalid number of arguments")
                return
            process_kill_command(args, host, port) 
        elif cmd == "status":
            if len(args) > 1:
                print_msg("Error: Invalid number of arguments")
                return
            process_status_command(args, host, port)     
        elif cmd == "list":
            if len(args) != 0:
                print_msg("Error: Invalid number of arguments")
                return
            process_list_command(args, host, port)
        elif cmd == "monit":
            if len(args) != 0:
                print_msg("Error: Invalid number of arguments")
                return
            process_monit_command(args, host, port)
            
    except ConnectionRefusedError:
        print_msg("Error: pypm is not running")
        
def process_monit_command(args, host, port):
    from .monit import App
    resp = send_command(const.CMD_LIST, args, host, port)
    if isdata(resp):
        strings = resp[1:].decode("utf-8").split("\x00\x00")
        if strings == ['']:
            total = None
        else:
            total = map(lambda l: l.split("\x00"), strings)
    app = App(host, port)
    if total is not None:
        for name, proc in total:
            app.add_process(name, proc)
    app.start()
        
def process_status_command(args, host, port):
    """Prints the status table for a given process/list of processes"""
    mem = process_mem_command(args, host, port)
    if mem is None:
        return
    if len(mem) == 0:
        print_msg("Warning: There are no processes being managed")
        return
    cpu = process_cpu_command(args, host, port)
    if cpu is None:
        return
    pid = process_pid_command(args, host, port)
    if pid is None:
        return
    uptime = process_uptime_command(args, host, port)
    if uptime is None:
        return
        
    lines = []
    for name in mem:
        memory = mem[name]
        if name in pid:
            active = f"{Fore.GREEN}active{Style.RESET_ALL}"
            p = pid[name]
            if p == -1:
                p = "N/A"
                active = f"{Fore.RED}stopped{Style.RESET_ALL}"
        else:
            p = "N/A"
            active = "N/A"
        if name in cpu:
            c = str(cpu[name])+"%"
        else:
            c = "N/A"
        if name in uptime:
            up = uptime[name]
        else:
            up = "N/A"
        
        lines.append([name, p, memory, c, up, active])
        
    header = ["Name", "PID", "Mem.", "CPU", "Uptime", "Status"]
    table = tt.to_string(
        lines,
        header=list(map(lambda c: color(c, Fore.CYAN), header)),
    )
    print(table)
        
def process_list_command(args, host, port):
    """List all managed processes"""
    resp = send_command(const.CMD_LIST, args, host, port)
    if isdata(resp):
        strings = resp[1:].decode("utf-8").split("\x00\x00")
        if strings == ['']:
            print_msg("Warning: There are no processes being managed")
        else:
            names, procs = zip(*map(lambda l: l.split("\x00"), strings))
            for i in range(len(names)):
                print_msg(f"* {names[i]} -> {procs[i]}")
    else:
        print_msg(resp[1:].decode())    
        
def process_mem_command(args, host, port):
    """Get the memory usage of a specific process/list of processes"""
    resp = send_command(const.CMD_GET_MEMORY, args, host, port)
    if isdata(resp):
        values = {}
        i = 0
        while i+1 < len(resp) and b"\x00" in resp[1+i:]:
            end = resp[1+i:].index(b"\x00")+1
            name = resp[1+i:i+end].decode()
            value = Size(struct.unpack("d", resp[1+i+end:i+end+9])[0])
            values[name] = value
            i += end+8
        return values
    else:
        print_msg(resp[1:].decode())
        return None
    
def process_cpu_command(args, host, port):
    """Get the cpu usage of a specific process/list of processes"""
    resp = send_command(const.CMD_GET_CPU, args, host, port)
    if isdata(resp):
        values = {}
        i = 0
        while i+1 < len(resp) and b"\x00" in resp[1+i:]:
            end = resp[1+i:].index(b"\x00")+1
            name = resp[1+i:i+end].decode()
            value = struct.unpack("d", resp[1+i+end:i+end+9])[0]
            values[name] = value
            i += end+8
        return values
    else:
        print_msg(resp[1:].decode())
        return None
    
def process_pid_command(args, host, port):
    """Get the PID of a specific process/list of processes"""
    resp = send_command(const.CMD_GET_PID, args, host, port)
    if isdata(resp):
        values = {}
        i = 0
        while i+1 < len(resp) and b"\x00" in resp[1+i:]:
            end = resp[1+i:].index(b"\x00")+1
            name = resp[1+i:i+end].decode()
            value = struct.unpack("i", resp[1+i+end:i+end+5])[0]
            values[name] = value
            i += end+4
        return values
    else:
        print_msg(resp[1:].decode())
        return None
    
def process_uptime_command(args, host, port):
    """Get the uptime of a specific process/list of processes"""
    resp = send_command(const.CMD_GET_UPTIME, args, host, port)
    if isdata(resp):
        values = {}
        i = 0
        while i+1 < len(resp) and b"\x00" in resp[1+i:]:
            end = resp[1+i:].index(b"\x00")+1
            name = resp[1+i:i+end].decode()
            i += end
            end = resp[1+i:].index(b"\x00")+1
            value = resp[1+i:i+end].decode()
            values[name] = value
            i += end
        return values
    else:
        print_msg(resp[1:].decode())
        return None
    
def process_stdout_command(args, host, port):
    """Gets the last 100 lines of output from the process"""
    resp = send_command(const.CMD_GET_STDOUT, args, host, port)
    if isdata(resp):
        return resp[1:].decode().split("\n")
    else:
        print_msg(resp[1:])
        
def process_stderr_command(args, host, port):
    """Gets the last 100 lines of output from the process"""
    resp = send_command(const.CMD_GET_STDERR, args, host, port)
    if isdata(resp):
        return resp[1:].decode().split("\n")
    else:
        print_msg(resp[1:])
        
def process_stop_command(args, host, port):
    """Closes the pypm server running on the given host"""
    resp = send_command(const.CMD_STOP, args, host, port)
    print_msg(resp[1:].decode())
        
def process_add_command(args, host, port):
    """Adds a new process to be managed"""
    name, command = args[:2]
    log_cpu = args[2] if len(args) >= 3 else "False"
    log_freq = args[3] if len(args) == 4 else "False"
    resp = send_command(const.CMD_ADD_PROCESS, 
                        [name, command, log_cpu, log_freq], 
                        host, port)
    print_msg(resp[1:].decode())
        
def process_restart_command(args, host, port):
    """Restarts a given process/list of processes"""
    resp = send_command(const.CMD_RESTART_PROCESS, args, host, port)
    print_msg(resp[1:].decode())
    
def process_start_command(args, host, port):
    """Starts a specific process/list of processes"""
    resp = send_command(const.CMD_START_PROCESS, args, host, port)
    print_msg(resp[1:].decode())
        
def process_kill_command(args, host, port):
    """Stops a specific process/list of processes"""
    resp = send_command(const.CMD_KILL_PROCESS, args, host, port)
    print_msg(resp[1:].decode())
        
def process_remove_command(args, host, port):
    """Removes (and stops) a process"""
    resp = send_command(const.CMD_REMOVE_PROCESS, args, host, port)
    print_msg(resp[1:].decode())

def send_command(cmd, args, host, port):
    string = ' '.join([cmd]+args)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((host, port))
    sock.sendall(string.encode("utf-8"))
    resp = b""
    while True:
        data = sock.recv(2048)
        if data == b"":
            break
        resp += data
        if len(data) < 2048:
            break
    sock.close()
    return resp
    
if __name__ == "__main__":
    import subprocess
    import sys
    
    from .pypm import main
    
    
    if len(sys.argv) >= 2:
        cmd = sys.argv[1]
        sys.argv.pop(1)
    else:
        print_msg(help_text)
        quit()
    
    if cmd == "init":
        
        argparser = get_start_parser()
        args, _ = argparser.parse_known_args()
        print_msg(f"Starting process manager on port {args.port}...")
        
        # ! This is only for debugging purposes. It makes it so the start 
        # ! command hangs and prints all output to the terminal window 
        # ! where it was called from
        # try:
        #     main(args.port, args.logdir, args.logfreq)
        # except socket.error:
        #     print_msg("Error: this port is already in use")
        
        # * This would be the actual production code
        
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            result = sock.connect_ex(("127.0.0.1", args.port))
            if result == 0:
                print_msg("Error: this port is already in use")
                quit()
        except socket.error:
            print_msg("Error: this port is already in use")
            quit()
        
        pid = subprocess.Popen([sys.executable, 
                                "-m", 
                                "pypm.pypm", 
                                str(args.port), 
                                str(args.logdir), 
                                str(args.logfreq)],
                shell=False, stdin=None, stdout=None, stderr=None,
                 close_fds=True, creationflags=0x00000008).pid
        print_msg(f"Started process manager on port {args.port} with the PID {pid}")
        
    elif cmd in commands:
        argparser = get_cmd_parser(cmd)
        args, _ = argparser.parse_known_args()
        for a in range(len(args.args)):
            arg = args.args[a]
            if " " in arg:
                args.args[a] = "'"+arg+"'"
        process_command(cmd, args.args, args.host, args.port)
    else:
        print_msg(help_text)



CMD_GET_MEMORY = "procmem"
CMD_GET_MEMORY_PERC = "procmemp"
CMD_GET_CPU = "proccpu"
CMD_ADD_PROCESS = "addproc"
CMD_KILL_PROCESS = "killproc"
CMD_START_PROCESS = "startproc"
CMD_RESTART_PROCESS = "reproc"
CMD_REMOVE_PROCESS = "remproc"
CMD_LIST = "list"
CMD_STOP = "stop"
CMD_GET_PID = "procpid"
CMD_GET_UPTIME = "procupt"
CMD_GET_STDOUT = "procstdout"
CMD_GET_STDERR = "porcstderr"

DATA_CODE = b"\x00"
MSG_CODE = b"\x01"


import logging
import os
import select
import shlex
import socket
import struct
import threading
import time

from . import constants as const
from .process import Process


def sbool(string):
    return True if string == "True" else False


# TODO: Add documentation
class ProcessManager:
    def __init__(self, port=8080, log_dir=None, log_frequency=30):
        self.port = port
        self.log_dir = log_dir
        self.log_frequency = log_frequency
        self._processes = []
        self._log_cpu = []
        self._log_memory = []
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._server_thread = None
        self._stop = False
        
    def add_process(self, process, log_cpu=False, log_memory=False):
        """Adds a process to be managed.

        Args:
            process (Process): The process
            log_cpu (bool, optional): True if CPU usage should be tracked. Defaults to False.
            log_memory (bool, optional): True if memory usage should be tracked. Defaults to False.

        Returns:
            bool: True if process wasn't already added
        """
        
        if process in self._processes:
            return False
        self._processes.append(process)
        if log_cpu and process not in self._log_cpu:
            self._log_cpu.append(process)
        if log_memory and process not in self._log_memory:
            self._log_memory.append(process) 
        return True
            
    def rem_process(self, process):
        """Removes a process"""
        self._processes.remove(process)
        if process in self._log_cpu:
            self._log_cpu.remove(process)
        if process in self._log_memory:
            self._log_memory.remove(process)
            
    def assert_logdir_exists(self):
        if self.log_dir is None:
            raise ValueError("Log directory wasn't specified")
        if not os.path.isdir(self.log_dir):
            os.mkdir(self.log_dir)
            
    def log_process_cpu(self, process):
        self.assert_logdir_exists()
        log_file = os.path.join(self.log_dir, process.name)
        with open(log_file+"_log_cpu", "ab") as file:
            file.write(struct.pack("d", process.get_cpu_perc()))
            
    def log_process_memory(self, process):
        self.assert_logdir_exists()
        log_file = os.path.join(self.log_dir, process.name)
        with open(log_file+"_log_mem", "ab") as file:
            file.write(struct.pack("d", process.get_mem_usage().bytes))
            
    def _process_command(self, command, sock):
        try:
            command = shlex.split(command)
            if len(command) == 0:
                sock.sendall(const.MSG_CODE+b"Error: Unrecognized command")  
            elif command[0] == const.CMD_GET_MEMORY:
                self._process_get_mem_cmd(command, sock)
            elif command[0] == const.CMD_GET_CPU:
                self._process_get_cpu_cmd(command, sock)
            elif command[0] == const.CMD_GET_PID:
                self._process_get_pid_cmd(command, sock)
            elif command[0] == const.CMD_GET_UPTIME:
                self._process_get_uptime_cmd(command, sock)
            elif command[0] == const.CMD_ADD_PROCESS:
                self._process_command_add_proc(command, sock)
            elif command[0] == const.CMD_RESTART_PROCESS:
                self._process_command_restart_proc(command, sock)
            elif command[0] == const.CMD_START_PROCESS:
                self._process_command_start_proc(command, sock)
            elif command[0] == const.CMD_STOP:
                self._process_command_stop(command, sock)
            elif command[0] == const.CMD_REMOVE_PROCESS:
                self._process_command_rem_proc(command, sock)
            elif command[0] == const.CMD_KILL_PROCESS:
                self._process_command_kill_proc(command, sock)
            elif command[0] == const.CMD_GET_STDERR:
                self._process_get_stderr(command, sock)
            elif command[0] == const.CMD_GET_STDOUT:
                self._process_get_stdout(command, sock)
            elif command[0] == const.CMD_LIST:
                self._process_list_cmd(command, sock)
            else:
                sock.sendall(const.MSG_CODE+b"Error: Unrecognized command") 
        except ConnectionResetError:
            pass
            
    def _process_get_stdout(self, command, sock):
        try:
            if len(command) == 2:
                name = command[1]
                out = None
                for process in self._processes:
                    if process.name == name:
                        out = process.stdout
                if out is None:
                    sock.sendall(const.MSG_CODE+b"Error: Couldn't find process '" + name.encode() + b"'")
                else:
                    sock.sendall(const.DATA_CODE+'\n'.join(out).encode())
            else:
                sock.sendall(const.MSG_CODE+b"Error: Invalid number of arguments")
        except Exception:
            sock.sendall(const.MSG_CODE+b"Error: Couldn't get stdout")
            
    def _process_get_stderr(self, command, sock):
        try:
            if len(command) == 2:
                name = command[1]
                err = None
                for process in self._processes:
                    if process.name == name:
                        err = process.stderr
                if err is None:
                    sock.sendall(const.MSG_CODE+b"Error: Couldn't find process '" + name.encode() + b"'")
                else:
                    sock.sendall(const.DATA_CODE+'\n'.join(err).encode())
            else:
                sock.sendall(const.MSG_CODE+b"Error: Invalid number of arguments")
        except Exception:
            sock.sendall(const.MSG_CODE+b"Error: Couldn't get stderr")
    
    def _process_list_cmd(self, command, sock):
        try:
            if len(command) == 1:
                ldata = lambda p: p.name + "\x00" + p.command
                data = '\x00\x00'.join(map(ldata, self._processes)).encode("utf-8")
                sock.sendall(const.DATA_CODE+data)
            else:
                sock.sendall(const.MSG_CODE+b"Error: Invalid number of arguments")
        except Exception:
            sock.sendall(const.MSG_CODE+b"Error: Couldn't get process list")
            
    def _process_get_uptime_cmd(self, command, sock):
        try:
            if 1 <= len(command) <= 2:
                if len(command) == 2:
                    name = command[1]
                    uptime = None
                    for process in self._processes:
                        if process.name == name:
                            uptime = str(process.uptime)
                            break 
                    if uptime is None:
                        sock.sendall(const.MSG_CODE+b"Error: Couldn't find process '" + name.encode() + b"'")
                    else:
                        sock.sendall(const.DATA_CODE+name.encode()+b"\x00"+uptime.encode()+b"\x00")
                else:
                    uptime = []
                    for process in self._processes:
                        uptime.append((process.name, str(process.uptime)))
                    message = []
                    for up in uptime:
                        message.append(up[0].encode()+b"\x00"+up[1].encode()+b"\x00")
                    sock.sendall(const.DATA_CODE+b"".join(message))
            else:
                sock.sendall(const.MSG_CODE+b"Error: Invalid number of arguments")
        except Exception:
            sock.sendall(const.MSG_CODE+b"Error: Couldn't get process uptime")
            
    def _process_get_mem_cmd(self, command, sock):
        try:
            if 1 <= len(command) <= 2:
                if len(command) == 2:
                    name = command[1]
                    memory = None
                    for process in self._processes:
                        if process.name == name:
                            memory = process.get_mem_usage().bytes   
                            break 
                    if memory is None:
                        sock.sendall(const.MSG_CODE+b"Error: Couldn't find process '" + name.encode() + b"'")
                    else:
                        sock.sendall(const.DATA_CODE+name.encode()+b"\x00"+struct.pack("d", memory))
                else:
                    memory = []
                    for process in self._processes:
                        memory.append((process.name, process.get_mem_usage().bytes))
                    message = []
                    for mem in memory:
                        message.append(mem[0].encode()+b"\x00"+struct.pack("d", mem[1]))
                    sock.sendall(const.DATA_CODE+b"".join(message))
            else:
                sock.sendall(const.MSG_CODE+b"Error: Invalid number of arguments")
        except Exception:
            sock.sendall(const.MSG_CODE+b"Error: Couldn't get process memory usage")
            
    def _process_get_pid_cmd(self, command, sock):
        try:
            if 1 <= len(command) <= 2:
                if len(command) == 2:
                    name = command[1]
                    pid = None
                    for process in self._processes:
                        if process.name == name:
                            pid = process.pid  
                            break 
                    if pid is None:
                        sock.sendall(const.MSG_CODE+b"Error: Couldn't find process '" + name.encode() + b"'")
                    else:
                        sock.sendall(const.DATA_CODE+name.encode()+b"\x00"+struct.pack("i", pid))
                else:
                    pid = []
                    for process in self._processes:
                        pid.append((process.name, process.pid))
                    message = []
                    for p in pid:
                        message.append(p[0].encode()+b"\x00"+struct.pack("i", p[1]))
                    sock.sendall(const.DATA_CODE+b"".join(message))
            else:
                sock.sendall(const.MSG_CODE+b"Error: Invalid number of arguments")
        except Exception:
            sock.sendall(const.MSG_CODE+b"Error: Couldn't get process PID")
            
    def _process_get_cpu_cmd(self, command, sock):
        try:
            if 1 <= len(command) <= 2:
                if len(command) == 2:
                    name = command[1]
                    cpu = None
                    for process in self._processes:
                        if process.name == name:
                            cpu = process.get_cpu_perc()   
                            break 
                    if cpu is None:
                        sock.sendall(const.MSG_CODE+b"Error: Couldn't find process '" + name.encode() + b"'")
                    else:
                        sock.sendall(const.DATA_CODE+name.encode()+b"\x00"+struct.pack("d", cpu))
                else:
                    cpu = []
                    for process in self._processes:
                        cpu.append((process.name, process.get_cpu_perc()))
                    message = []
                    for c in cpu:
                        message.append(c[0].encode()+b"\x00"+struct.pack("d", c[1]))
                    sock.sendall(const.DATA_CODE+b"".join(message))
            else:
                sock.sendall(const.MSG_CODE+b"Error: Invalid number of arguments")
        except Exception:
            sock.sendall(const.MSG_CODE+b"Error: Couldn't get process CPU usage")
            
    def _process_command_add_proc(self, command, sock):
        try:
            if len(command) == 5:
                name, cmd, log_cpu, log_freq = command[1:]
                process = Process(name, cmd)
                if self.add_process(process, sbool(log_cpu), sbool(log_freq)):
                    sock.sendall(const.MSG_CODE+b"Successfully added process '" + name.encode() + b"'")
                else:
                    sock.sendall(const.MSG_CODE+b"Error: There is already a process named '" + name.encode() + b"'")
            else:
                sock.sendall(const.MSG_CODE+b"Error: Invalid number of arguments")
        except Exception:
            sock.sendall(const.MSG_CODE+b"Error: Couldn't add process")
            
    def _process_command_restart_proc(self, command, sock):
        try:
            if not (1 <= len(command) <= 2):
                sock.sendall(const.MSG_CODE+b"Error: Invalid number of arguments")
                return
            if len(command) == 2:
                name = command[1]
                process = None
                for proc in self._processes:
                    if proc.name == name:
                        process = proc
                        break 
                if process is None:
                    sock.sendall(const.MSG_CODE+b"Error: Couldn't find process '" + name.encode() + b"'")
                    return
                if process.active:
                    process.kill()
                process.start(True)
                sock.sendall(const.MSG_CODE+b"Successfully restarted process '" + name.encode() + b"'")
            else:
                if len(self._processes) == 0:
                    sock.sendall(const.MSG_CODE+b"Warning: No processes to restart")
                    return
                c = 0
                for process in self._processes:
                    try:
                        if process.active:
                            process.kill()
                        process.start(True)
                        c += 1
                    except Exception:
                        pass
                if c == 0:
                    sock.sendall(const.MSG_CODE+b"Warning: No processes were restarted")
                else:
                    total = str(c).encode()
                    length = str(len(self._processes)).encode()
                    sock.sendall(const.MSG_CODE+b"Restarted " + total + b" out of " + length + b" processes")
                
        except Exception:
            sock.sendall(const.MSG_CODE+b"Error: Couldn't restart process")
    
    def _process_command_start_proc(self, command, sock):
        try:
            if not (1 <= len(command) <= 2):
                sock.sendall(const.MSG_CODE+b"Error: Invalid number of arguments")
                return
            if len(command) == 2:
                name = command[1]
                process = None
                for proc in self._processes:
                    if proc.name == name:
                        process = proc
                        break 
                if process is None:
                    sock.sendall(const.MSG_CODE+b"Error: Couldn't find process '" + name.encode() + b"'")
                    return
                if process.active:
                    sock.sendall(const.MSG_CODE+b"Warning: Process was already running, so nothing was done")
                else:
                    process.start(True)
                    sock.sendall(const.MSG_CODE+b"Successfully started process '" + name.encode() + b"'")
            else:
                if len(self._processes) == 0:
                    sock.sendall(const.MSG_CODE+b"Warning: No processes to start")
                    return
                c = 0
                for process in self._processes:
                    if not process.active:
                        try:
                            process.start(True)
                            c += 1
                        except Exception:
                            pass
                if c == 0:
                    sock.sendall(const.MSG_CODE+b"Warning: No processes were started")
                else:
                    total = str(c).encode()
                    length = str(len(self._processes)).encode()
                    sock.sendall(const.MSG_CODE+b"Started " + total + b" out of " + length + b" processes")
                
        except Exception:
            sock.sendall(const.MSG_CODE+b"Error: Couldn't start process")
            
    def _process_command_rem_proc(self, command, sock):
        try:
            if len(command) != 2:
                sock.sendall(const.MSG_CODE+b"Error: Invalid number of arguments")
                return
            name = command[1]
            process = None
            for proc in self._processes:
                if proc.name == name:
                    process = proc
                    break 
            if process is None:
                sock.sendall(const.MSG_CODE+b"Error: Couldn't find process '" + name.encode() + b"'")
                return
            if process.active:
                process.kill()
            self.rem_process(process)
            sock.sendall(const.MSG_CODE+b"Successfully removed process '" + name.encode() + b"'")
            
        except Exception:
            sock.sendall(const.MSG_CODE+b"Error: Couldn't remove process")
            
    def _process_command_kill_proc(self, command, sock):
        try:
            if len(command) != 2:
                sock.sendall(const.MSG_CODE+b"Error: Invalid number of arguments")
                return
            name = command[1]
            process = None
            for proc in self._processes:
                if proc.name == name:
                    process = proc
                    break 
            if process is None:
                sock.sendall(const.MSG_CODE+b"Error: Couldn't find process '" + name.encode() + b"'")
                return
            if process.active:
                process.kill()
                sock.sendall(const.MSG_CODE+b"Successfully killed process '" + name.encode() + b"'")
            else:
                sock.sendall(const.MSG_CODE+b"Error: Process '" + name.encode() + b"' is not active")
            
        except Exception:
            sock.sendall(const.MSG_CODE+b"Error: Couldn't kill process")
            
    def _process_command_stop(self, command, sock):
        host = socket.gethostname().encode()
        port = str(self.port).encode()
        sock.sendall(const.MSG_CODE+b"Stopped pypm running on " + host + b":" + port)
        self._stop = True
        
    @property
    def has_active_processes(self):
        return len(list(filter(lambda p: p.active, self._processes))) >= 1
    
    @property
    def log_period(self):
        return 60 / self.log_frequency
    
    def start(self):
        self._socket.bind(("localhost", self.port))
        for process in self._processes:
            process.start(True)
        self.main_loop()
        
    def server_loop(self):
        self._socket.listen()
        while not self._stop:
            try:
                sock, _ = self._socket.accept()
                command = sock.recv(2048).decode("utf-8")
                self._process_command(command, sock)
                sock.close()
            except ConnectionResetError:
                pass
        
    def main_loop(self):
        try:
            start = time.time()
            self._server_thread = threading.Thread(target=self.server_loop)
            self._server_thread.start()
            while not self._stop:
                if time.time() - start > self.log_period:
                    
                    start = time.time()
                    for process in self._processes:
                        if process in self._log_memory:
                            self.log_process_memory(process)
                        if process in self._log_cpu:
                            self.log_process_cpu(process)
                        if process.active:
                            process.process_stdout()
                            process.process_stderr()
        except KeyboardInterrupt:    
            pass
        finally:
            self._stop = True
            
            # * In case the server_loop hasn't stopped yet, prevent
            # * socket.accept() from hanging by connecting
            try:
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.setblocking(0)
                s.settimeout(0.5)
                s.connect(("localhost", self.port))
                s.sendall(b" ")
                ready = select.select([s], [], [], 0.6)  # See if data is available
                if ready[0]:
                    s.recv(1024)
                s.close()
            except:
                pass
            
            self._socket.close()
            for process in self._processes:
                if process.active:
                    process.kill()



import curses
import threading
import time
import traceback

from .__main__ import (process_cpu_command, process_mem_command,
                       process_pid_command, process_stderr_command,
                       process_stdout_command, process_uptime_command)
from .units import Size, Time

CTRL_Z = 26
CTRL_C = 3
K_UP = 450
K_DOWN = 456
K_LEFT = 452
K_RIGHT = 454
K_RETURN = 10
K_ESCAPE = 27
K_SPACE = 32
DISPLAY = {
    "command": "Command",
    "pid": "PID",
    "mem": "Memory Usage",
    "cpu": "CPU Usage",
    "uptime": "Uptime"
}

def wrap(string, width):
    c = 0
    while c < len(string):
        yield string[c:c+width]
        c += width

def pad(string, size):
    return string[:size] + " "*max(0, size-len(string))

class App:
    def __init__(self, host, port):
        self._host = host
        self._port = port
        self._processes = {}
        self._selected_proc = 0
        self._log_offset = 0
        self._log_mode = "stderr"
        self._proc_offset = 0
        self._selected = 0
        self._screen = None
        self._topleftwin = None
        self._toprightwin = None
        self._botrightwin = None
        self._should_update = {
            "topleft": False,
            "topright": False,
            "botright": False            
        }
        self.WHITE = None
        self.YELLOW = None
        self.GREEN = None
        self.BLUE = None
        self.SELECT = None
        self.RED = None
        
        self._stop = False
        
    def update_info(self):
        try:
            start = time.time()
            while not self._stop:
                if len(self._processes) > 0:
                    keys = list(self._processes.keys())
                    proc = keys[self._selected_proc]
                    if time.time()-start > 1:
                        mem = process_mem_command([proc], self._host, self._port)
                        if mem is None:
                            break
                        cpu = process_cpu_command([proc], self._host, self._port)
                        if cpu is None:
                            break
                        uptime = process_uptime_command([proc], self._host, self._port)
                        if uptime is None:
                            break
                        pid = process_pid_command([proc], self._host, self._port)
                        if pid is None:
                            break
                        stdout = process_stdout_command([proc], self._host, self._port)
                        stderr = process_stderr_command([proc], self._host, self._port)
                        self._processes[proc]["pid"] = pid[proc] if pid[proc] != -1 else "N/A"
                        self._processes[proc]["uptime"] = uptime[proc]
                        self._processes[proc]["mem"] = mem[proc]
                        self._processes[proc]["cpu"] = str(cpu[proc])+"%"
                        self._processes[proc]["logs"]["stdout"] = stdout
                        self._processes[proc]["logs"]["stderr"] = stderr
                        start = time.time()
                    self.schedule_update(["botright", "topright"])
        except Exception:
            traceback.print_exc()
            pass
        finally:
            self._stop = True
            
    def add_process(self, name, command):
        self._processes[name]={
            "command": command,
            "pid": "N/A",
            "uptime": "0s",
            "mem": "0.0B",
            "cpu": "0.0%",
            "logs": {
                "stdout": [],
                "stderr": []
            }
        }
        
    def schedule_update(self, screens=[]):
        if screens == []:
            self._should_update.update(
                zip(self._should_update.keys(), [True]*len(self._should_update))
            )
        else:
            self._should_update.update(
                zip(screens, [True]*len(screens))
            )
        
    def start(self):
        thread = threading.Thread(target=self.update_info)
        thread.setDaemon(True)
        thread.start()
        curses.wrapper(self.main_loop)
        
    def setup(self):
        if curses.LINES < 17 or curses.COLS < 54:
            print("Your terminal isn't big enough")
            return False
        curses.curs_set(False)
        self._screen.nodelay(True)
        lsize = (curses.LINES-1, curses.COLS//3)
        self._topleftwin = curses.newwin(*lsize, 0, 0)
        height = curses.LINES-9
        width = curses.COLS-lsize[1]
        self._toprightwin = curses.newwin(height, width, 0, lsize[1])
        self._botrightwin = curses.newwin(curses.LINES-height-1, width, height, lsize[1])
        
        curses.use_default_colors()
        curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLACK)
        curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)
        curses.init_pair(3, curses.COLOR_BLUE, curses.COLOR_BLACK)
        curses.init_pair(4, curses.COLOR_WHITE, curses.COLOR_BLUE)
        curses.init_pair(5, curses.COLOR_RED, curses.COLOR_BLACK)
        curses.init_pair(6, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        
        self.WHITE = curses.color_pair(1)
        self.GREEN = curses.color_pair(2)
        self.BLUE = curses.color_pair(3)
        self.SELECT = curses.color_pair(4)
        self.RED = curses.color_pair(5)
        self.YELLOW = curses.color_pair(6)
        return True
    
    def get_log_lines(self, ltype):
        max_x = self._toprightwin.getmaxyx()[1]-4
        keys = list(self._processes.keys())
        proc_name = keys[self._selected_proc]
        proc = self._processes[proc_name]
        c = 0
        for line in proc["logs"][ltype]:
            for sub in wrap(line, max_x):
                c += 1
        return c
        
    def update_topleftwin(self):
        self._topleftwin.clear()
        if self._selected == 0:
            self._topleftwin.attron(self.BLUE)
        self._topleftwin.box()
        self._topleftwin.attroff(self.BLUE)
        self._topleftwin.addstr(0, 2, " Process List ")
        max_y , max_x = self._topleftwin.getmaxyx()
        max_x -= 4
        max_y -= 4
        for i in range(len(self._processes)):
            if i-self._proc_offset < 0:
                continue
            if i-self._proc_offset >= max_y:
                break
            proc = list(self._processes.keys())[i]
            if i == self._selected_proc:
                self._topleftwin.attron(self.SELECT)
            self._topleftwin.addstr(2+i-self._proc_offset, 2, pad(proc, max_x))
            self._topleftwin.attroff(self.SELECT)
        if len(self._processes) != self._proc_offset+max_y:
            if len(self._processes) > max_y:
                self._topleftwin.attron(self.YELLOW)
                self._topleftwin.addstr(max_y+2, 2, "...".center(max_x))
                self._topleftwin.attroff(self.YELLOW)
        self._topleftwin.refresh()
        
    def update_toprightwin(self):
        self._toprightwin.clear()
        if self._selected == 1:
            self._toprightwin.attron(self.BLUE)
        self._toprightwin.box()
        self._toprightwin.attroff(self.BLUE)
        self._toprightwin.addstr(0, 2, f" Logs ({self._log_mode}) ")
        self._toprightwin.move(2, 0)
        max_y , max_x = self._toprightwin.getmaxyx()
        max_x -= 4
        max_y -= 3
        if len(self._processes) > 0:
            keys = list(self._processes.keys())
            proc_name = keys[self._selected_proc]
            proc = self._processes[proc_name]
            c = self.get_log_lines(self._log_mode)
            start = max(0, c-max_y)
            l = 2-start-self._log_offset
            for line in proc["logs"][self._log_mode]:
                for sub in wrap(line, max_x):
                    if l >= 2:
                        if "warning" in sub.lower():
                            self._toprightwin.attron(self.YELLOW)
                        elif "error" in sub.lower() or "critical" in sub.lower():
                            self._toprightwin.attron(self.RED)
                        self._toprightwin.addstr(l, 2, sub)
                        self._toprightwin.attroff(self.YELLOW|self.RED)
                    l += 1
                    if l > max_y:
                        break
                if l > max_y:
                    break
        self._toprightwin.refresh()
        
    def update_botrightwin(self):
        self._botrightwin.clear()
        if self._selected == 2:
            self._botrightwin.attron(self.BLUE)
        self._botrightwin.box()
        self._botrightwin.attroff(self.BLUE)
        self._botrightwin.addstr(0, 2, " Status: ")
        max_y , max_x = self._botrightwin.getmaxyx()
        max_x -= 4
        max_y -= 3
        if len(self._processes) > 0:
            keys = list(self._processes.keys())
            proc_name = keys[self._selected_proc]
            proc = self._processes[proc_name]
            if proc["pid"] == "N/A":
                self._botrightwin.attron(self.RED)
            else:
                self._botrightwin.attron(self.GREEN)
            self._botrightwin.addstr(0, 10, " ⬤ ")
            self._botrightwin.attroff(self.RED|self.GREEN)
            
            self._botrightwin.addstr(1, 2, f"Name: {pad(proc_name, max_x-5)}")
            for i in range(len(proc)-1):
                if i >= max_y:
                    break
                attr = list(proc.keys())[i]
                self._botrightwin.addstr(2+i, 2, f"{DISPLAY[attr]}: {proc[attr]}")
        self._botrightwin.refresh()
        
    def update_proc_offset(self):
        max_y = self._topleftwin.getmaxyx()[0]-4
        if len(self._processes) < max_y:
            self._proc_offset = 0
        else:
            if self._selected_proc-self._proc_offset >= max_y:
                self._proc_offset = self._selected_proc-max_y+1
            elif self._selected_proc == 0:
                self._proc_offset = 0
            elif self._selected_proc-self._proc_offset <= 0:
                self._proc_offset = self._selected_proc
        
    def main_loop(self, screen):
        self._screen = screen
        if not self.setup():
            return
        self._screen.clear()
        self._botrightwin.clear()
        self._toprightwin.clear()
        self._topleftwin.clear()
        string = "Press Ctrl+C to exit | Use [SPACE] to change log mode".center(curses.COLS-1)
        self._screen.addstr(curses.LINES-1, 0, string)
        self.schedule_update()
        
        while not self._stop:
            char = self._screen.getch()
            if char != -1:
                if char == CTRL_C or char == CTRL_Z:
                    break
                elif char == K_SPACE:
                    self._log_offset = 0
                    if self._selected == 1:
                        if self._log_mode == "stderr":
                            self._log_mode = "stdout"
                        else:
                            self._log_mode = "stderr"
                elif char == K_RIGHT or char == curses.KEY_RIGHT:
                    self._selected = (self._selected + 1)%3
                    self.schedule_update()
                elif char == K_LEFT or char == curses.KEY_LEFT:
                    self._selected = (self._selected - 1)%3
                    self.schedule_update()
                elif char == K_UP or char == curses.KEY_UP:
                    if self._selected == 0:
                        if len(self._processes) != 0:
                            self._selected_proc -= 1
                            self._selected_proc %= len(self._processes)
                            self.update_proc_offset()
                            self.schedule_update(["topleft"])
                    elif self._selected == 1:
                        if len(self._processes) != 0:
                            keys = list(self._processes.keys())
                            proc_name = keys[self._selected_proc]
                            proc = self._processes[proc_name]
                            lines = self.get_log_lines(self._log_mode)
                            max_y = self._toprightwin.getmaxyx()[0] - 3
                            if self._log_offset > -(lines - max_y):
                                self._log_offset -= 1
                                self.schedule_update(["topright"])
                elif char == K_DOWN or char == curses.KEY_DOWN:
                    if self._selected == 0:
                        if len(self._processes) != 0:
                            self._selected_proc += 1
                            self._selected_proc %= len(self._processes)
                            self.update_proc_offset()
                            self.schedule_update(["topleft"])
                    elif self._selected == 1:
                        if len(self._processes) != 0:
                            keys = list(self._processes.keys())
                            proc_name = keys[self._selected_proc]
                            proc = self._processes[proc_name]
                            lines = self.get_log_lines(self._log_mode)
                            if self._log_offset < 0:
                                self._log_offset += 1
                                self.schedule_update(["topright"])
                
            if self._should_update["topleft"]:
                self.update_topleftwin()
                self._should_update["topleft"] = False
            if self._should_update["topright"]:
                self.update_toprightwin()
                self._should_update["topright"] = False  
            if self._should_update["botright"]:
                self.update_botrightwin()
                self._should_update["botright"] = False  
                

if __name__ == "__main__":
    app = App()
    app.start()



import sys

from .manager import ProcessManager


def main(port=8080, log_dir=None, log_freq=30):
    if log_dir == "None":
        log_dir = None
    pm = ProcessManager(port=port, log_dir=log_dir, log_frequency=log_freq)
    pm.start()
    
if __name__ == "__main__":
    main(int(sys.argv[1]), sys.argv[2], float(sys.argv[3]))


import datetime
import subprocess
import tempfile
import threading

import psutil

from .units import Size, Time


class Process:
    def __init__(self, name, command):
        self.max_buff_size = 100
        self.name = name
        self._command = command
        self._process = None
        self._start = Time(0)
        self._cpu_usage = 0
        self._thread = None
        self._outstream = None
        self._errstream = None
        self._outbuff = []
        self._errbuff = []
        
    def __eq__(self, other):
        return isinstance(other, Process) and other.name == self.name
        
    def start(self, pipe=False):
        if self.active:
            raise OSError("Process is already running")
        self._start = datetime.datetime.now()
        if pipe:
            self._outstream = tempfile.TemporaryFile()
            self._errstream = tempfile.TemporaryFile()
            self._process = subprocess.Popen(self._command.split(),
                                             stdout=self._outstream,
                                             stderr=self._errstream)
        else:
            self._process = subprocess.Popen(self._command.split())
            
    @property
    def stdout(self):
        return self._outbuff
    
    @property
    def stderr(self):
        return self._errbuff
            
    def process_stdout(self):
        new = self.get_stdout().decode()
        if new != "":
            self._outbuff = self._outbuff + new.split("\n")
        while len(self._outbuff) > self.max_buff_size:
            self._outbuff.pop(0)
            
    def process_stderr(self):
        new = self.get_stderr().decode()
        if new != "":
            self._errbuff = self._errbuff + new.split("\n")  
        while len(self._errbuff) > self.max_buff_size:
            self._errbuff.pop(0)
        
    def get_stdout(self):
        self._outstream.flush()
        self._outstream.seek(0)
        r = self._outstream.read()
        if b"\n" in r:
            nl = r.rindex(b"\n")
            self._outstream.truncate(0)
            self._outstream.write(r[nl+1:])
            return r[:nl]
        else:
            return b""
    
    def get_stderr(self):
        self._errstream.flush()
        self._errstream.seek(0)
        r = self._errstream.read()
        if b"\n" in r:
            nl = r.rindex(b"\n")
            self._errstream.truncate(0)
            self._errstream.write(r[nl+1:])
            return r[:nl]
        else:
            return b""
        
    def kill(self):
        self._start = Time(0)
        self._process.kill()
        self._outstream.close()
        self._errstream.close()
        
    def update_cpu(self):
        try:
            self._cpu_usage = psutil.Process(self.pid).cpu_percent(0.5) / psutil.cpu_count()
        except psutil.NoSuchProcess:
            pass
        
    @property
    def command(self):
        return self._command
        
    @property
    def active(self):
        if self._process is None:
            return False
        return self._process.poll() is None
    
    @property 
    def pid(self):
        if self.active:
            return self._process.pid
        return -1
    
    @property
    def uptime(self):
        if self.active:
            return Time(datetime.datetime.now()-self._start)
        else:
            return Time(0)
    
    def get_mem_usage(self):
        if self.active:
            return Size(psutil.Process(self.pid).memory_info().vms)
        else:
            return Size(0)
    
    def get_mem_perc(self):
        if self.active:
            return psutil.Process(self.pid).memory_percent("vms")
        else:
            return 0
    
    def get_cpu_perc(self):
        if self.active:
            if self._thread is None or not self._thread.is_alive():
                self._thread = threading.Thread(target=self.update_cpu)
                self._thread.setDaemon(True)
                self._thread.start()
            return self._cpu_usage
        else:
            return 0



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
    
with open("requirements.txt", "r") as fh:
    requirements = fh.read().split("\n")

setuptools.setup(
    name="python-pm",
    version="0.0.1",
    author="Francisco Rodrigues",
    author_email="francisco.rodrigues0908@gmail.com",
    description="A simple, python-based process manager",
    install_requires=requirements,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ArmindoFlores/pypm",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
